#![allow(non_camel_case_types)]

use std::os::fd::{AsRawFd, OwnedFd, RawFd};
use std::ffi::{CStr, OsString, c_void};
use libc;
use crate::gpu::drm_i915::DrmDeviceNode;

use super::native;
use std::fs;
use std::fs::File;

#[derive(Debug)]
pub struct DrmVersion {
    pub version_major: i32,
    pub version_minor: i32,
    pub version_patchlevel: i32,
    pub name: OsString,
    pub date: OsString,
    pub desc: OsString,
}

pub fn get_drm_version(fd: RawFd) -> Option<DrmVersion> {
    unsafe {
        let name_vec = libc::malloc(128);
        libc::memset(name_vec, 0, 128);
        let date_vec = libc::malloc(128);
        libc::memset(date_vec, 0, 128);
        let desc_vec = libc::malloc(128);
        libc::memset(desc_vec, 0, 128);
        
        let mut version = native::drm_version { 
            version_major: 0,
            version_minor: 0,
            version_patchlevel: 0,
            name: name_vec as *mut i8,
            date: date_vec as *mut i8,
            desc: desc_vec as *mut i8,
            name_len: 128,
            date_len: 128,
            desc_len: 128,
        };
        let res_val = libc::ioctl(fd, native::DRM_IOCTL_VERSION, &mut version) == 0;
        let res = if res_val {
            let name = CStr::from_ptr(version.name).to_str().expect("Failed to convert name");
            let date = CStr::from_ptr(version.date).to_str().expect("Failed to convert date");
            let desc = CStr::from_ptr(version.desc).to_str().expect("Failed to convert desc");
            let drm_version = DrmVersion {
                version_major: version.version_major,
                version_minor: version.version_minor,
                version_patchlevel: version.version_patchlevel,
                name: OsString::from(name.to_string()),
                date: OsString::from(date.to_string()),
                desc: OsString::from(desc.to_string()),
            };
            Some(drm_version)
        } else {None};
        res
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum DrmI915EngineClass {
    Render, Copy, Video, VideoEnhance, Compute, Invalid
}

#[derive(Debug)]
pub struct DrmI915EngineClassInstance {
    pub engine_class: DrmI915EngineClass,
    pub engine_instance: u16,
}

pub type DrmI915EngineInfoFlags = u64;

#[derive(Debug)]
pub struct DrmI915EngineInfo {
    pub engine: DrmI915EngineClassInstance,
    pub logical_instance: u16,
    pub flags: DrmI915EngineInfoFlags,
    pub capabilities: u64,
}

fn engine_class_from_u16(engine_class: u16) -> DrmI915EngineClass {
    match engine_class {
        0 => DrmI915EngineClass::Render,
        1 => DrmI915EngineClass::Copy,
        2 => DrmI915EngineClass::Video,
        3 => DrmI915EngineClass::VideoEnhance,
        4 => DrmI915EngineClass::Compute,
        _ => DrmI915EngineClass::Invalid,
    }
}

fn engine_class_instance_from_native(engine_class_instance: native::i915_engine_class_instance) -> DrmI915EngineClassInstance {
    DrmI915EngineClassInstance {
        engine_class: engine_class_from_u16(engine_class_instance.engine_class),
        engine_instance: engine_class_instance.engine_instance,
    }
}

fn engine_info_from_native(engine_info: native::drm_i915_engine_info) -> DrmI915EngineInfo {
    DrmI915EngineInfo {
        engine: engine_class_instance_from_native(engine_info.engine),
        logical_instance: engine_info.logical_instance,
        flags: engine_info.flags,
        capabilities: engine_info.capabilities,
    }
}

impl DrmI915EngineInfo {
    pub fn to_native(&self) -> native::drm_i915_engine_info {
        native::drm_i915_engine_info {
            engine: native::i915_engine_class_instance {
                engine_class: match self.engine.engine_class {
                    DrmI915EngineClass::Render => 0,
                    DrmI915EngineClass::Copy => 1,
                    DrmI915EngineClass::Video => 2,
                    DrmI915EngineClass::VideoEnhance => 3,
                    DrmI915EngineClass::Compute => 4,
                    DrmI915EngineClass::Invalid => 0xFFFF,
                },
                engine_instance: self.engine.engine_instance,
            },
            logical_instance: self.logical_instance,
            flags: self.flags,
            capabilities: self.capabilities,
            rsvd0: 0,
            rsvd1: [0; 3],
            rsvd2: [0; 3],
        }
    }
    pub fn from_native(engine_info: native::drm_i915_engine_info) -> DrmI915EngineInfo {
        engine_info_from_native(engine_info)
    }
}

pub fn get_engines(fd: RawFd) -> Option<Vec<DrmI915EngineInfo>> {
    unsafe {
        let mut query_item = native::drm_i915_query_item {
            query_id: native::DRM_I915_QUERY_ENGINE_INFO as u64,
            length: 0,
            data_ptr: core::ptr::null_mut::<c_void>() as u64,
            flags: 0,
        };
        let mut query = native::drm_i915_query {
            items_ptr: (&mut query_item) as *mut native::drm_i915_query_item as u64,
            num_items: 1,
            flags: 0,
        };
        let res = libc::ioctl(fd.as_raw_fd(), native::DRM_IOCTL_I915_QUERY, &mut query);
        if res != 0 {
            return None;
        }
        let engine_info = libc::malloc(query_item.length as usize)
            as *mut native::drm_i915_query_engine_info;
        libc::memset(engine_info as *mut c_void, 0, query_item.length as usize);
        query_item.data_ptr = engine_info as u64;
        let res = libc::ioctl(fd.as_raw_fd(), native::DRM_IOCTL_I915_QUERY, &mut query);
        if res != 0 {
            return None;
        }
        let engine_data = engine_info.as_mut().unwrap();
        let engines = engine_data.engines.as_mut_slice(engine_data.num_engines as usize);
        let mut engines_res: Vec<DrmI915EngineInfo> = Vec::new();
        for i in 0..engine_data.num_engines {
            let engine = DrmI915EngineInfo::from_native(engines[i as usize]);
            engines_res.push(engine);            
        }
        Some(engines_res)
    }
}

pub fn get_param(fd: RawFd, param_id: i32) -> Option<i32> {
    unsafe {
        let mut value: i32 = 0;
        let mut param = native::drm_i915_getparam {
            param: param_id,
            value: &mut value as *mut i32
        };
        let res = libc::ioctl(fd, native::DRM_IOCTL_I915_GETPARAM, &mut param);
        if res != 0 {
            return None;
        }
        Some(value)
    }
}

pub fn find_i915_node() -> Option<DrmDeviceNode> {
    let inodes = fs::read_dir("/dev/dri").ok()?;
    for inode in inodes {
        let inode = inode.ok()?;
        let path = inode.path();
        let path = path.to_str()?;
        if path.starts_with("/dev/dri/renderD") {
            let file = File::open(path).unwrap();
            let dev = get_drm_version(file.as_raw_fd());
            if dev.is_some() && dev.unwrap().name.to_str().unwrap() == "i915" {
                return Some(DrmDeviceNode {
                    fd: OwnedFd::from(file),
                    path: Some(OsString::from(path)),
                });
            }
        }
    }
    None
}

pub fn find_i915_fd() -> Option<RawFd> {
    return find_i915_node().map(|node| node.fd.as_raw_fd());
}